package glk;

/* 
   Model.java
   Ported from Andrew Plotkin's model.c
*/

/* This is a simple model of a text adventure which uses the Glk API.
    It shows how to input a line of text, display results, maintain a
    status window, write to a transcript file, and so on. */

/* This isn't as pure as model.c; the java.lang methods are used freely */

public class Model extends GlkProxy implements GlkProgram {

    /* The story, status, and quote windows. */
    winid_t mainwin = null;
    winid_t statuswin = null;
    winid_t quotewin = null;

    /* A file reference for the transcript file. */
    frefid_t scriptref = null;
    /* A stream for the transcript file, when it's open. */
    strid_t scriptstr = null;

    /* Your location. This determines what appears in the status line. */
    int current_room; 

    /* A flag indicating whether you should look around. */
    int need_look; 
    
    /* The glk_main() function is called by the Glk system; it's the main entry
       point for your program.  It is passed a pointer to the glk library */
    public void glk_main(Glk glklib)
    {
	glk = glklib;
	/* Open the main window. */
	mainwin = glk_window_open(null, 0, 0, wintype_TextBuffer, 1);
	if (mainwin == null) {
	    /* It's possible that the main window failed to open. There's
	       nothing we can do without it, so exit. */
	    return; 
	}
	
	/* Set the current output stream to print to it. */
	glk_set_window(mainwin);
    
	/* Open a second window: a text grid, above the main window, three lines
	   high. It is possible that this will fail also, but we accept that. */
	statuswin = glk_window_open(mainwin, winmethod_Above | winmethod_Fixed, 
					3, wintype_TextGrid, 0);
	
	/* The third window, quotewin, isn't opened immediately. We'll do
	   that in verb_quote(). */

	glk_put_string("Model Glk Program\nAn Interactive Model Glk Program\n");
	glk_put_string("By Andrew Plotkin.\nRelease 7.\n");
	glk_put_string("Type \"help\" for a list of commands.\n");

	current_room = 0; /* Set initial location. */
	need_look = TRUE;
    
	byte [] commandbuf = new byte[256];
	while (true) {
	    int gotline, len;
	    event_t ev = new event_t();
	    
	    draw_statuswin();
        
	    if (need_look == TRUE) {
		need_look = FALSE;
		glk_put_string("\n");
		glk_set_style(style_Subheader);
		if (current_room == 0)
		    glk_put_string("The Room\n");
		else
		    glk_put_string("A Different Room\n");
		glk_set_style(style_Normal);
		glk_put_string("You're in a room of some sort.\n");
	    }
        
	    glk_put_string("\n>");
	    /* We request up to 255 characters. The buffer can hold 256, but we
	       are going to stick a null character at the end, so we have to
	       leave room for that. Note that the Glk library does *not*
	       put on that null character. */
	    glk_request_line_event(mainwin, commandbuf, 255, 0);
        
	    gotline = FALSE;
	    do {
        
		/* Grab an event. */
		glk_select(ev);
		
		switch (ev.type) {
		    
                case evtype_LineInput:
                    if (ev.win == mainwin) {
                        gotline = TRUE;
                        /* Really the event can *only* be from mainwin,
			   because we never request line input from the
			   status window. But we do a paranoia test,
			   because commandbuf is only filled if the line
			   event comes from the mainwin request. If the
			   line event comes from anywhere else, we ignore
			   it. */
                    }
                    break;
                    
                case evtype_Arrange:
                    /* Windows have changed size, so we have to redraw the
		       status window. */
                    draw_statuswin();
                    break;
		}
	    }
	    while (gotline == FALSE);
	    
	    /* commandbuf now contains a line of input from the main window.
	       You would now run your parser and do something with it. */
	    
	    /* First, if there's a blockquote window open, let's close it. 
	       This ensures that quotes remain visible for exactly one
	       command. */
	    if (quotewin != null) {
		glk_window_close(quotewin, null);
		quotewin = null;
	    }
	    
	    /* first convert the command buffer to a string */
	    len = ev.val1;
	    String cmd = new String(commandbuf, 0, len, "ISO-8859-1");
	    
	    /* Then squash to lower-case. */
	    cmd = cmd.toLowerCase();
	    
	    /* Then trim whitespace before and after. */
	    cmd = cmd.trim();
        
        
	    /* cmd now points to a nice null-terminated string. We'll do the
	       simplest possible parsing. */
	    if (cmd.equals("")) {
		glk_put_string("Excuse me?\n");
	    }
	    else if (cmd.equals("help")) {
		verb_help();
	    }
	    else if (cmd.equals("move")) {
		verb_move();
	    }
	    else if (cmd.equals("jump")) {
		verb_jump();
	    }
	    else if (cmd.equals("yada")) {
		verb_yada();
	    }
	    else if (cmd.equals("quote")) {
		verb_quote();
	    }
	    else if (cmd.equals("quit")) {
		verb_quit();
	    }
	    else if (cmd.equals("save")) {
		verb_save();
	    }
	    else if (cmd.equals("restore")) {
		verb_restore();
	    }
	    else if (cmd.equals("script")) {
		verb_script();
	    }
	    else if (cmd.equals("unscript")) {
		verb_unscript();
	    }
	    else {
		glk_put_string("I don't understand the command \"");
		glk_put_string(cmd);
		glk_put_string("\".\n");
	    }
	}
    }
    
    void draw_statuswin()
    {
	String roomname;
	int width, height;
	
	if (statuswin == null) {
	    /* It is possible that the window was not successfully 
	       created. If that's the case, don't try to draw it. */
	    return;
	}
	
	if (current_room == 0)
	    roomname = "The Room";
	else
	    roomname = "A Different Room";
	
	glk_set_window(statuswin);
	glk_window_clear(statuswin);
    
	glk_size size = glk_window_get_size(statuswin);
	width = size.width;
	height = size.height;
    
	/* Print the room name, centered. */
	glk_window_move_cursor(statuswin, (width - roomname.length()) / 2, 1);
	glk_put_string(roomname);
    
	/* Draw a decorative compass rose in the upper right. */
	glk_window_move_cursor(statuswin, width - 3, 0);
	glk_put_string("\\|/");
	glk_window_move_cursor(statuswin, width - 3, 1);
	glk_put_string("-*-");
	glk_window_move_cursor(statuswin, width - 3, 2);
	glk_put_string("/|\\");
    
	glk_set_window(mainwin);
    }
    
    int yes_or_no()
    {
	byte commandbuf[] = new byte[256];
	int gotline, len;
	event_t ev = new event_t();
    
	draw_statuswin();
	
	/* This loop is identical to the main command loop in glk_main(). */
	
	while (true) {
	    glk_request_line_event(mainwin, commandbuf, 255, 0);
        
	    gotline = FALSE;
	    do {
		
		glk_select(ev);
		
		switch (ev.type) {
                case evtype_LineInput:
                    if (ev.win == mainwin) {
                        gotline = TRUE;
                    }
                    break;
                    
                case evtype_Arrange:
                    draw_statuswin();
                    break;
		}
	    }
	    while (gotline == FALSE);	    

	    len = ev.val1;
	    String cmd = new String(commandbuf, 0, len, "ISO-8859-1");
	    cmd = cmd.trim();
	    if (cmd.length() > 0) {
		char cx0 = cmd.charAt(0);
		if (cx0 == 'y' || cx0 == 'Y')
		    return TRUE;
		if (cx0 == 'n' || cx0 == 'N')
		    return FALSE;
	    }

	    glk_put_string("Please enter \"yes\" or \"no\": ");
	}
	
    }

    void verb_help()
    {
	glk_put_string("This model only understands the following commands:\n");
	glk_put_string("HELP: Display this list.\n");
	glk_put_string("JUMP: A verb which just prints some text.\n");
	glk_put_string("YADA: A verb which prints a very long stream of text.\n");
	glk_put_string("MOVE: A verb which prints some text, and also changes the status line display.\n");
	glk_put_string("QUOTE: A verb which displays a block quote in a temporary third window.\n");
	glk_put_string("SCRIPT: Turn on transcripting, so that output will be echoed to a text file.\n");
	glk_put_string("UNSCRIPT: Turn off transcripting.\n");
	glk_put_string("SAVE: Write fake data to a save file.\n");
	glk_put_string("RESTORE: Read it back in.\n");
	glk_put_string("QUIT: Quit and exit.\n");
    }

    void verb_jump()
    {
	glk_put_string("You jump on the fruit, spotlessly.\n");
    }

    static final int NUMWORDS = 13;
    
    String wordcaplist[] = {
	"Ga", "Bo", "Wa", "Mu", "Bi", "Fo", "Za", "Mo", "Ra", "Po",
	"Ha", "Ni", "Na"
    };

    String wordlist[] = {
        "figgle", "wob", "shim", "fleb", "moobosh", "fonk", "wabble",
	"gazoon", "ting", "floo", "zonk", "loof", "lob",
    };
    int wcount1 = 0;
    int wcount2 = 0;
    int wstep = 1;
    int jx = 0;
    void verb_yada()
    {
	/* This is a goofy (and overly ornate) way to print a long paragraph. 
	   It just shows off line wrapping in the Glk implementation. */
	int ix;
	int first = TRUE;
    
	for (ix=0; ix<85; ix++) {
	    if (ix > 0) {
		glk_put_string(" ");
	    }
	    
	    if (first == TRUE) {
		glk_put_string(wordcaplist[(ix / 17) % NUMWORDS]);
		first = FALSE;
	    }
        
	    glk_put_string(wordlist[jx]);
	    jx = (jx + wstep) % NUMWORDS;
	    wcount1++;
	    if (wcount1 >= NUMWORDS) {
		wcount1 = 0;
		wstep++;
		wcount2++;
		if (wcount2 >= NUMWORDS-2) {
		    wcount2 = 0;
		    wstep = 1;
		}
	    }
	    
	    if ((ix % 17) == 16) {
		glk_put_string(".");
		first = TRUE;
	    }
	}
	
	glk_put_char('\n');
    }

    void verb_quote()
    {
	glk_put_string("Someone quotes some poetry.\n");

	/* Open a third window, or clear it if it's already open. Actually,
	   since quotewin is closed right after line input, we know it
	   can't be open. But better safe, etc. */
	if (quotewin == null) {
	    /* A five-line window above the main window, fixed size. */
	    quotewin = glk_window_open(mainwin, winmethod_Above | winmethod_Fixed, 
				       5, wintype_TextBuffer, 0);
	    if (quotewin == null) {
		/* It's possible the quotewin couldn't be opened. In that
		   case, just give up. */
		return;
	    }
	}
	else {
	    glk_window_clear(quotewin);
	}
	
	/* Print some quote. */
	glk_set_window(quotewin);
	glk_set_style(style_BlockQuote);
	glk_put_string("Tomorrow probably never rose or set\n" +
			   "Or went out and bought cheese, or anything like that\n" +
			   "And anyway, what light through yonder quote box breaks\n" +
			   "Handle to my hand?\n");
	glk_put_string("              -- Fred\n");
	
	glk_set_window(mainwin);
    }
    
    void verb_move()
    {
	current_room = (current_room+1) % 2;
	need_look = TRUE;
	
	glk_put_string("You walk for a while.\n");
    }

    void verb_quit()
    {
	glk_put_string("Are you sure you want to quit? ");
	if (yes_or_no() == TRUE) {
	    glk_put_string("Thanks for playing.\n");
	    glk_exit();
	    /* glk_exit() actually stops the process; it does not return. */
	}
    }
    
    void verb_script()
    {
	if (scriptstr != null) {
	    glk_put_string("Scripting is already on.\n");
	    return;
	}
	
	/* If we've turned on scripting before, use the same file reference; 
	   otherwise, prompt the player for a file. */
	if (scriptref == null) {
	    scriptref = glk_fileref_create_by_prompt(
						     fileusage_Transcript | fileusage_TextMode, 
						     filemode_WriteAppend, 0);
	    if (scriptref == null) {
		glk_put_string("Unable to place script file.\n");
		return;
	    }
	}
	
	/* Open the file. */
	scriptstr = glk_stream_open_file(scriptref, filemode_WriteAppend, 0);
	if (scriptstr == null) {
	    glk_put_string("Unable to write to script file.\n");
	    return;
	}
	glk_put_string("Scripting on.\n");
	glk_window_set_echo_stream(mainwin, scriptstr);
	glk_put_string_stream(scriptstr, 
				  "This is the beginning of a transcript.\n");
    }
    
    void verb_unscript()
    {
	if (scriptstr == null) {
	    glk_put_string("Scripting is already off.\n");
	    return;
	}
    
	/* Close the file. */
	glk_put_string_stream(scriptstr, 
				  "This is the end of a transcript.\n\n");
	glk_stream_close(scriptstr, null);
	glk_put_string("Scripting off.\n");
	scriptstr = null;
    }

    void verb_save()
    {
	int ix;
	frefid_t saveref;
	strid_t savestr;
	
	saveref = glk_fileref_create_by_prompt(
						   fileusage_SavedGame | fileusage_BinaryMode, 
						   filemode_Write, 0);
	if (saveref == null) {
	    glk_put_string("Unable to place save file.\n");
	    return;
	}
	
	savestr = glk_stream_open_file(saveref, filemode_Write, 0);
	if (savestr == null) {
	    glk_put_string("Unable to write to save file.\n");
	    glk_fileref_destroy(saveref);
	    return;
	}
	
	glk_fileref_destroy(saveref); /* We're done with the file ref now. */
    
	/* Write some binary data. */
	for (ix=0; ix<256; ix++) {
	    glk_put_char_stream(savestr, (char)ix);
	}
    
	glk_stream_close(savestr, null);
	
	glk_put_string("Game saved.\n");
    }

    void verb_restore()
    {
	int ix;
	int err;
	int ch;
	frefid_t saveref;
	strid_t savestr;
	
	saveref = glk_fileref_create_by_prompt(
						   fileusage_SavedGame | fileusage_BinaryMode, 
						   filemode_Read, 0);
	if (saveref == null) {
	    glk_put_string("Unable to find save file.\n");
	    return;
	}
	
	savestr = glk_stream_open_file(saveref, filemode_Read, 0);
	if (savestr == null) {
	    glk_put_string("Unable to read from save file.\n");
	    glk_fileref_destroy(saveref);
	    return;
	}
	
	glk_fileref_destroy(saveref); /* We're done with the file ref now. */
	
	/* Read some binary data. */
	err = FALSE;
	
	for (ix=0; ix<256; ix++) {
	    ch = glk_get_char_stream(savestr);
	    if (ch == (-1)) {
		glk_put_string("Unexpected end of file.\n");
		err = TRUE;
		break;
	    }
	    if (ch != ix) {
		glk_put_string("This does not appear to be a valid saved game.\n");
		err = TRUE;
		break;
	    }
	}
	
	glk_stream_close(savestr, null);
	
	if (err == TRUE) {
	    glk_put_string("Failed.\n");
	    return;
	}
	
	glk_put_string("Game restored.\n");
    }
}
    
